import datetime
import json
import math

import numpy as np
import pandas as pd
import pkg_resources

journal_mapped = "../tables/journal_mapped.json"
journal_mapped_path = pkg_resources.resource_filename(__name__, journal_mapped)


class JournalChecker:
    def __init__(self, journal: dict = None, work_day: int = 11) -> None:
        if journal is not None:
            self.journal = journal
        else:
            file = open(journal_mapped_path, encoding="utf8")
            journal_data = json.load(file)
            self.journal = journal_data
        self.work_day = work_day

    def validate_resources(
        self, pool: list, res_name: str, model_dataset: pd.DataFrame, index: datetime
    ):
        res_volume = model_dataset.loc[index, res_name]
        sum_res = 0
        for w in pool:
            work_volume = model_dataset.loc[index, w]
            if w.split("_act_fact")[0] not in self.journal:
                return "grey", 0, 0
            else:
                if (
                    res_name.split("_res_fact")[0]
                    in self.journal[w.split("_act_fact")[0].split("_act_fact")[0]][
                        "resources"
                    ]
                ):
                    model = self.journal[w.split("_act_fact")[0].split("_act_fact")[0]][
                        "resources"
                    ][res_name.split("_res_fact")[0]]
                    sum_res += (
                        (work_volume * model["number"])
                        / model["productivity"]
                        / self.work_day
                    )
        q1 = sum_res - sum_res * 0.5
        q99 = sum_res + sum_res * 0.5
        if (res_volume < q1) | (res_volume > q99):
            return "red", q1, q99
        else:
            return "green", q1, q99

    def validate_time(self, c_act: list, df_wind_model: pd.DataFrame, index: int):
        if c_act[0].split("_act_fact")[0] not in self.journal:
            return "grey", 0, 0, 0
        else:
            act_volume = df_wind_model.loc[index, c_act[0]]
            res_for_act = []
            avg_productivity = 0
            for r in self.journal[c_act[0].split("_act_fact")[0]]["resources"]:
                if r + "_res_fact" in df_wind_model.columns:
                    number = self.journal[c_act[0].split("_act_fact")[0]]["resources"][
                        r
                    ]["number"]
                    value_to_append = (
                        df_wind_model.loc[index, r + "_res_fact"]
                    ) / number

                    if value_to_append != 0:
                        res_for_act.append(value_to_append)

                    avg_productivity += self.journal[c_act[0].split("_act_fact")[0]][
                        "resources"
                    ][r]["productivity"]

            avg_productivity /= len(res_for_act)
            min_res = np.min(res_for_act) * 11
            prod = avg_productivity * min_res
            time = act_volume / prod
            q1 = time - time * 0.5
            q99 = time + time * 0.5
            act_time = df_wind_model.loc[
                index, c_act[0].split("_act_fact")[0] + "_real_time_act"
            ]
            if (act_time < q1) | (act_time > q99):
                return "red", q1, q99, prod
            else:
                return "green", q1, q99, prod

    def process_key_resources(self, c_act, r, ci, model_dataset, i):
        key = str(c_act) + " " + r + " " + ci
        work_volume = model_dataset.loc[i, ci]
        res_volume = model_dataset.loc[i, r]
        power = self._check_volume_power(work_volume)
        step = 10 ** (power - 1)
        low_bound = work_volume - step
        up_bound = work_volume + step
        X = list(np.linspace(low_bound, work_volume, 15)) + list(
            np.linspace(work_volume, up_bound, 15)
        )
        if (
            r.split("_res_fact")[0]
            not in self.journal[ci.split("_act_fact")[0]]["resources"]
        ):
            Y = [0] * len(X)
            blue_points = {
                "x": list(X),
                "y": list(Y),
            }
            black_points = {
                "x": [work_volume],
                "y": [0],
            }
            star = {
                "x": work_volume,
                "y": res_volume,
            }
            return key, blue_points, black_points, star
        else:
            model = self.journal[ci.split("_act_fact")[0]]["resources"][
                r.split("_res_fact")[0]
            ]
            Y = []
            for xi in X:
                Y.append((xi * model["number"]) / model["productivity"] / self.work_day)
            blue_points = {
                "x": list(X),
                "y": list(Y),
            }
            black_points = {
                "x": [work_volume],
                "y": [
                    (work_volume * model["number"])
                    / model["productivity"]
                    / self.work_day
                ],
            }
            star = {
                "x": work_volume,
                "y": res_volume,
            }
            return key, blue_points, black_points, star

    def process_key_time(self, c_act, df_wind_model, i, prod):
        work_volume = df_wind_model.loc[i, c_act[0]]
        power = self._check_volume_power(work_volume)
        step = 10 ** (power - 1)
        low_bound = work_volume - step
        up_bound = work_volume + step
        X = list(np.linspace(low_bound, work_volume, 15)) + list(
            np.linspace(work_volume, up_bound, 15)
        )
        Y = []
        for xi in X:
            Y.append(xi / prod)
        blue_points = {
            "x": list(X),
            "y": list(Y),
        }
        black_points = {
            "x": [work_volume],
            "y": [work_volume / prod],
        }
        star = {
            "x": work_volume,
            "y": df_wind_model.loc[
                i, c_act[0].split("_act_fact")[0] + "_real_time_act"
            ],
        }
        return blue_points, black_points, star

    @staticmethod
    def _check_volume_power(volume):
        power = math.log10(volume)
        return math.floor(power) if power < 0 else math.ceil(power)



import datetime

import pandas as pd

from stairsval.core.dataset_processors.BaseDataset import BaseDataset


class PlanDataset(BaseDataset):
    def __init__(self, ksg_data):
        self.ksg_data = ksg_data

    def collect(self):
        model_dataset = pd.DataFrame()
        all_descendants = []

        activity_ids = set()
        start = datetime.datetime.today()
        for i, w in enumerate(self.ksg_data["schedule"]["works"]):
            activity_id = w["id"]
            if activity_id in activity_ids:
                raise Exception(
                    f"ERROR: Duplicated activity_id detected: {activity_id}"
                )
            activity_ids.add(activity_id)

            current_descendant_activities = self.ksg_data["wg"]["nodes"][i].get(
                "parent_edges", []
            )
            all_descendants.extend(current_descendant_activities)

            name = w["display_name"] + "_act_fact"
            days = w["start_end_time"][1]["value"] - w["start_end_time"][0]["value"]
            vol = float(w["volume"])
            res_data = dict()
            for r in w["workers"]:
                res_data[r["name"]] = r["_count"]
            if not res_data:
                raise Exception(
                    'ERROR: One or more activities has no resources. Check "workers" fields'
                )
            vol_per = vol / days
            start_d = start + datetime.timedelta(days=w["start_end_time"][0]["value"])
            end_d = start + datetime.timedelta(days=w["start_end_time"][1]["value"])
            delta = datetime.timedelta(days=1)
            while start_d <= end_d:
                formatted_date = start_d.strftime("%d.%m.%Y")
                model_dataset.loc[formatted_date, name] = vol_per
                for k in res_data.keys():
                    model_dataset.loc[formatted_date, k + "_res_fact"] = res_data[k]
                start_d += delta
        model_dataset.fillna(0, inplace=True)
        return model_dataset

    def get_act_names(self):
        act = []
        for w in self.ksg_data["schedule"]["works"]:
            act.append(w["display_name"])
        return act

    def get_res_names(self):
        res = []
        for w in self.ksg_data["schedule"]["works"]:
            for r in w["workers"]:
                if r["name"] not in res:
                    res.append(r["name"])
        return res

    def get_pools(self):
        model_dataset = pd.DataFrame()
        start = datetime.datetime.today()
        for i, w in enumerate(self.ksg_data["schedule"]["works"]):
            name = w["display_name"]
            # Keep start_d and end_d as datetime objects
            start_d = start + datetime.timedelta(days=w["start_end_time"][0]["value"])
            end_d = start + datetime.timedelta(days=w["start_end_time"][1]["value"])
            vol = float(w["volume"])
            # Subtract the datetime objects
            days = (end_d - start_d).days + 1
            vol_per = vol / days
            delta = datetime.timedelta(days=1)
            while start_d <= end_d:
                formatted_date = start_d.strftime(
                    "%d.%m.%Y"
                )  # Convert to string format here
                model_dataset.loc[formatted_date, name] = vol_per
                start_d += delta
        model_dataset.fillna(0, inplace=True)
        work_pools = []
        for i in model_dataset.index:
            pool = []
            for c in model_dataset.columns:
                if model_dataset.loc[i, c] != 0:
                    pool.append(c)
            if pool not in work_pools:
                work_pools.append(pool)

        return work_pools



import numpy as np
import pandas as pd


class Aggregator:
    def __init__(self, brave_crit_value: float = 0.45):
        self.CRIT_VALUE = brave_crit_value

    @staticmethod
    def get_all_seq_statistic(start_end_data, model_data):
        works_dict = {
            act["id"]: act["display_name"] for act in model_data["schedule"]["works"]
        }
        pairs_labels = {
            (act["display_name"], works_dict[el[0]]): el[2]
            for i, act in enumerate(model_data["schedule"]["works"])
            if model_data["wg"]["nodes"][i]["parent_edges"]
            for el in model_data["wg"]["nodes"][i]["parent_edges"]
        }

        pairs = list(pairs_labels.keys())
        if not pairs:
            return pd.DataFrame(), pd.DataFrame(), 0, 0

        freq_dict = {}

        def check_dates(row):
            s1 = row["first_day_x"]
            f1 = row["last_day_x"]
            s2 = row["first_day_y"]
            f2 = row["last_day_y"]

            fs = int(f1 < s2)
            ss = int(s1 == s2)
            ff = int(f1 == f2)
            mix = int(s2 < f1)
            total = fs + ss + ff + mix

            return pd.Series(
                [fs, ss, ff, mix, total], index=["FS", "SS", "FF", "FFS", "count"]
            )

        for w1, w2 in pairs:
            ind1 = start_end_data.loc[(start_end_data["granular_smr_name"] == w1)]
            ind2 = start_end_data.loc[(start_end_data["granular_smr_name"] == w2)]
            merged_data = pd.merge(
                ind1, ind2, how="inner", on="upper_works", suffixes=("_x", "_y")
            )
            counts = pd.Series(
                [0, 0, 0, 0, 0], index=["FS", "SS", "FF", "FFS", "count"]
            )
            if merged_data.shape[0] != 0:
                counts = merged_data.apply(check_dates, axis=1)

            total_fs = counts["FS"].sum()
            total_ss = counts["SS"].sum()
            total_ff = counts["FF"].sum()
            total_ffs = counts["FFS"].sum()
            total_count = counts["count"].sum()

            if total_count > 0:
                freq_dict[w1, w2] = {
                    "count": total_count,
                    "FS": total_fs,
                    "SS": total_ss,
                    "FF": total_ff,
                    "FFS": total_ffs,
                }

        bar_records = []
        color_records = []
        links = ["FFS", "FS", "SS", "FF"]

        for i, ((w1, w2), label) in enumerate(pairs_labels.items()):
            pair_data = freq_dict.get((w1, w2), {})
            total_count = pair_data.get("count", 0)
            if total_count == 0:
                color_record = {
                    "Work name 1": w1,
                    "Work name 2": w2,
                    "color": "grey",
                }
                color_records.append(color_record)
                continue

            mix_val = pair_data.get("FFS", 0) / total_count
            fs_val = pair_data.get("FS", 0) / total_count
            ss_val = pair_data.get("SS", 0) / total_count
            ff_val = pair_data.get("FF", 0) / total_count
            links_perc = [mix_val, fs_val, ss_val, ff_val]
            max_label = links[np.argmax(links_perc)]

            bar_record = {
                "Work name 1": w1,
                "Work name 2": w2,
                "Plan connection": label,
                "FFS": mix_val * 100,
                "FS": fs_val * 100,
                "SS": ss_val * 100,
                "FF": ff_val * 100,
            }
            bar_records.append(bar_record)

            color_record = {
                "Work name 1": w1,
                "Work name 2": w2,
                "color": "green" if max_label == label else "red",
            }
            color_records.append(color_record)

        df_bar = pd.DataFrame.from_records(bar_records)
        df_color = pd.DataFrame.from_records(color_records)
        if df_color.empty:
            norm_perc = 0
            not_perc = 100
        else:
            perc_dict = df_color["color"].value_counts().to_dict()
            norm_perc = (perc_dict.get("green", 0) / df_color.shape[0]) * 100
            not_perc = (perc_dict.get("grey", 0) / df_color.shape[0]) * 100

        return df_bar, df_color, norm_perc, not_perc

    def get_res_ved_stat(self, proximity_model, ksg_for_val_data, plan_type):
        df_stat = pd.DataFrame(columns=["Work", "Resource", "Resource label"])

        def evaluate_resource(row):
            df_row = pd.DataFrame(columns=["Work", "Resource", "Resource label"])
            if plan_type == "gpn":
                res = row["workers"]
                act_name = row["display_name"]
                for i, r in enumerate(res):
                    df_row.loc[i, "Work"] = act_name
                    df_row.loc[i, "Resource"] = r["name"]
                    df_row.loc[i, "Resource label"] = None
                    if (
                        f"{act_name}_act_fact" in proximity_model.columns
                        and f"{r['name']}_res_fact" in proximity_model.index
                    ):
                        df_row.loc[i, "Resource label"] = (
                            "green"
                            if proximity_model.loc[
                                r["name"] + "_res_fact",
                                act_name + "_act_fact",
                            ]
                            >= self.CRIT_VALUE
                            else "red"
                        )
            elif plan_type == "s7":
                res = row["workers"]
                act_name = row["display_name"]
                if ":" in act_name:
                    name = act_name.replace(":", "_")
                    model = proximity_model[1].get_defect_res_probability(
                        evidence={"new_class_child": name}
                    )
                else:
                    model = proximity_model[0].get_res_probability(
                        evidence={"event_name_parent": act_name}
                    )
                for i, r in enumerate(res):
                    df_row.loc[i, "Work"] = act_name
                    df_row.loc[i, "Resource"] = r["name"]
                    df_row.loc[i, "Resource label"] = None
                    if len(model) != 0:
                        df_row.loc[i, "Resource label"] = (
                            "green" if model[r["name"]] >= self.CRIT_VALUE else "red"
                        )
            return df_row

        for work in ksg_for_val_data["schedule"]["works"]:
            df_result = evaluate_resource(work)
            df_stat = pd.concat([df_stat, df_result], ignore_index=True)

        not_grey = df_stat[df_stat["Resource label"] is not None]
        not_perc = (len(df_stat) - len(not_grey)) / len(df_stat) * 100
        norm_perc = (
            not_grey["Resource label"].value_counts().to_dict().get("green", 0)
            / len(not_grey)
        ) * 100

        return df_stat, not_perc, norm_perc



from datetime import datetime


class ExpertMetricsEstimator:
    def __init__(self, project_data):
        self.project = project_data

    def _get_all_descendant_acts(self, index=2):
        return [
            edge[index]
            for node in self.project["wg"]["nodes"]
            for edge in node["parent_edges"]
        ]

    def _calculate_percentage(self, condition, index=1):
        all_descendant_acts = self._get_all_descendant_acts(index)
        matching_count = sum(1 for desc in all_descendant_acts if condition(desc))
        return matching_count / len(all_descendant_acts) if all_descendant_acts else 0

    def follower_preds(self):
        descendant_percentage = self._calculate_percentage(
            lambda x: x not in self._get_all_descendant_acts(0), 0
        )
        return True if descendant_percentage <= 0.05 else False

    def outrun(self):
        fss_percentage = self._calculate_percentage(lambda x: x == "FFS")
        return True if fss_percentage <= 0.05 else False

    def finish_start(self):
        fs_ffs_percentage = self._calculate_percentage(lambda x: x in ["FS", "FFS"])
        return True if fs_ffs_percentage >= 0.85 else False

    def others(self):
        other_percentage = self._calculate_percentage(lambda x: x not in ["FS", "FFS"])
        return True if other_percentage <= 0.15 else False

    def crit_index(self):
        if "plan_deadline" not in self.project:
            return None

        acts_start_date = datetime.strptime(
            self.project["works"][0]["start_end_time"][0].split()[0], "%Y-%m-%d"
        )
        acts_deadline_date = datetime.strptime(
            self.project["plan_deadline"].split()[0], "%Y-%m-%d"
        )
        acts_end_date = datetime.strptime(
            self.project["works"][-1]["start_end_time"][-1].split()[0], "%Y-%m-%d"
        )
        ldeadline = (acts_deadline_date - acts_start_date).days
        lcp = (acts_end_date - acts_start_date).days
        index = ldeadline / lcp
        return True if index >= 1 else False

    def calculate_metrics(self):
        return {
            "follower_preds": self.follower_preds(),
            "outrun": self.outrun(),
            "finish_start": self.finish_start(),
            "others": self.others(),
            "crit_index": self.crit_index(),
        }

    def calculate_formal_metrics(self):
        # Filter out work names "start of projects" and "start" in schedule
        relevant_works = [
            work
            for work in self.project["schedule"]["works"]
            if work.get("name") not in ["start of projects", "start"]
        ]

        # Filter out work names "start of projects" and "start" in work units
        relevant_nodes = [
            node
            for node in self.project["wg"]["nodes"]
            if node.get("work_unit", {}).get("name")
            not in ["start of projects", "start"]
        ]

        # Check if all relevant works have non-empty 'workers' fields
        all_have_workers = all(
            "workers" in work and work["workers"] for work in relevant_works
        )

        # Check if all relevant work units have non-empty 'volume' fields
        all_have_volume = all(
            "work_unit" in node
            and "volume" in node["work_unit"]
            and node["work_unit"]["volume"]
            for node in relevant_nodes
        )

        # Check if at least one relevant work unit has a non-empty 'parent_edges' field
        at_least_one_has_parents = any(
            "work_unit" in node
            and "parent_edges" in node["work_unit"]
            and node["work_unit"]["parent_edges"]
            for node in relevant_nodes
        )

        return {
            "all_have_workers": all_have_workers,
            "all_have_volume": all_have_volume,
            "at_least_one_has_parents": at_least_one_has_parents,
        }



import pandas as pd
import pkg_resources
from rich.console import Console
from stairs_mro.networks.DefectRes import DefectRes
from stairs_mro.networks.WorkResNet import WorkResNet

from stairsval.core.aggregator import Aggregator
from stairsval.core.dataset_processors.S7KSG import S7KSG
from stairsval.core.expert_metrics_estimator import ExpertMetricsEstimator
from stairsval.validation_checks.ResourcesChecker import ResourcesChecker
from stairsval.validation_checks.TimeChecker import TimeChecker
from stairsval.validators.BaseValidator import BaseValidator

res_work = "../tables/work_res_data_no_nan_new2.csv"
res_defect = "../tables//work_defect_res_data_no_nan_new2.csv"
res_work_path = pkg_resources.resource_filename(__name__, res_work)
res_defect_path = pkg_resources.resource_filename(__name__, res_defect)


class S7PlanValidator:
    def __init__(self, plan, work_res_model: dict, defect_model: dict):
        self.console = Console()
        self.ksg_data = S7KSG(ksg_data=plan).collect()
        self.aggregator = Aggregator()
        self.time_validator = TimeChecker()
        self.resources_validator = None
        self.work_res_model = work_res_model
        self.defect_model = defect_model

    def validate(self):
        pass

    def common_validate(self):
        val_df_res_work = pd.read_csv(res_work_path)
        val_df_res_defect = pd.read_csv(res_defect_path)

        (
            df_validation_table_res,
            fig_dict_res,
            norm_perc_res_val,
            not_perc_res_val,
        ) = ResourcesChecker(res=[]).common_validation(
            self.ksg_data, (val_df_res_work, val_df_res_defect), plan_type="s7"
        )

        model_work = WorkResNet(structure=[])
        model_work.load(self.work_res_model)

        model_defect = DefectRes(structure=[])
        model_defect.load(self.defect_model)

        proximity_model = (model_work, model_defect)
        df_vedom, not_perc_vedom, norm_perc_vedom = self.aggregator.get_res_ved_stat(
            proximity_model, self.ksg_data, plan_type="s7"
        )
        (
            df_validation_table_time,
            fig_dict_time,
            norm_perc_time,
            not_perc_time,
        ) = TimeChecker().common_validation(
            self.ksg_data, (val_df_res_work, val_df_res_defect), "s7"
        )
        return (
            df_validation_table_res,
            fig_dict_res,
            norm_perc_res_val,
            not_perc_res_val,
            df_vedom,
            not_perc_vedom,
            norm_perc_vedom,
            df_validation_table_time,
            fig_dict_time,
            norm_perc_time,
            not_perc_time,
        )


class S7Validator(BaseValidator):

    """Validator for S7 plans."""

    def __init__(self, project_ksg, work_res_model: dict, defect_model: dict):
        super().__init__(project_ksg)
        self.plan_type = "s7"
        self.work_res_model = work_res_model
        self.defect_model = defect_model

    def specific_validation(self):
        return self.common_validation()

    def common_validation(self, cut_to_n_works: int = None):
        if cut_to_n_works:
            self._trim_plan_to_n_works(cut_to_n_works)
        (
            df_validation_table_res,
            fig_dict_res,
            norm_perc_res_val,
            not_perc_res_val,
            df_vedom,
            not_perc_vedom,
            norm_perc_vedom,
            df_validation_table_time,
            fig_dict_time,
            norm_perc_time,
            not_perc_time,
        ) = S7PlanValidator(
            plan=self.project_ksg,
            work_res_model=self.work_res_model,
            defect_model=self.defect_model,
        ).common_validate()

        work_res_stat = dict()
        work_res_stat["Percentage of Normal Resource Volume Values"] = round(
            norm_perc_res_val
        )
        work_res_stat["Percentage of Atypical Resource Volume Values"] = 100 - round(
            norm_perc_res_val
        )
        work_res_stat["Percentage of Normal Resources According to Statements"] = round(
            norm_perc_vedom
        )
        work_res_stat[
            "Percentage of Atypical Resources According to Statements"
        ] = 100 - round(norm_perc_vedom)
        work_res_stat["Percentage of Normal Values Across All Resources"] = round(
            (
                work_res_stat["Percentage of Normal Resource Volume Values"]
                + work_res_stat[
                    "Percentage of Normal Resources According to Statements"
                ]
            )
            / 2
        )
        work_res_stat["Percentage of Atypical Values Across All Resources"] = (
            100 - work_res_stat["Percentage of Normal Values Across All Resources"]
        )
        work_res_stat["Percentage of Normal Work Time Values"] = round(norm_perc_time)
        work_res_stat["Percentage of Atypical Work Time Values"] = 100 - round(
            norm_perc_time
        )
        work_res_stat["Percentage of Normal Values Across All Works"] = round(
            norm_perc_time
        )
        work_res_stat["Percentage of Atypical Values Across All Works"] = 100 - round(
            norm_perc_time
        )
        work_res_stat["Percentage of Normal Plan Values"] = round(
            (
                work_res_stat["Percentage of Normal Values Across All Resources"]
                + work_res_stat["Percentage of Normal Values Across All Works"]
            )
            / 2
        )
        work_res_stat["Percentage of Atypical Plan Values"] = (
            100 - work_res_stat["Percentage of Normal Plan Values"]
        )
        work_res_stat["Percentage of Plan Values Not Covered by Validation"] = round(
            (not_perc_res_val + not_perc_vedom + not_perc_time) / 3
        )
        work_res_stat["Percentage of Plan Values Covered by Validation"] = 100 - round(
            (not_perc_res_val + not_perc_vedom + not_perc_time) / 3
        )

        result_dict = dict()
        result_dict["Common validation"] = {
            "Resources Validation Table": df_validation_table_res,
            "Data for Resource Charts": fig_dict_res,
            "Resource Journal": df_vedom,
            "Work Time Validation Table": df_validation_table_time,
            "Data for Time Charts": fig_dict_time,
            "Final Plan Statistics": work_res_stat,
        }
        return result_dict

    def calculate_expert_metrics(self):
        metrics_calculator = ExpertMetricsEstimator(self.project_ksg)
        metrics = metrics_calculator.calculate_metrics()
        formal_metrics = metrics_calculator.calculate_formal_metrics()
        return metrics, formal_metrics



from enum import Enum, auto

from stairsval.validators.GPNValidator import GPNValidator
from stairsval.validators.S7Validator import S7Validator


class PlanType(Enum):
    GPN = auto()
    S7 = auto()


class ValidatorFactory:
    @staticmethod
    def create_validator(plan_type, *args, **kwargs):
        if not isinstance(plan_type, PlanType):
            raise ValueError("Invalid plan type provided!")

        if plan_type == PlanType.GPN:
            return GPNValidator(*args, **kwargs)
        elif plan_type == PlanType.S7:
            return S7Validator(*args, **kwargs)
        else:
            raise ValueError(f"No validator defined for plan type {plan_type}")



from abc import ABC, abstractmethod


class BaseValidator(ABC):
    """Abstract Base Validator class."""

    def __init__(self, project_ksg):
        self.project_ksg = project_ksg
        self._preprocess_project()

    @abstractmethod
    def common_validation(self, *args, **kwargs):
        """Common validation logic for all plans."""
        pass

    @abstractmethod
    def specific_validation(self):
        """Method to be overridden for plan-specific validation."""
        pass

    def _preprocess_project(self):
        # Names to be removed
        names_to_remove = {"start of project", "start"}

        # Remove works with specified names from the schedule
        self.project_ksg["schedule"]["works"] = [
            work
            for work in self.project_ksg["schedule"]["works"]
            if work.get("name") not in names_to_remove
        ]

        # Get the IDs of the removed works
        removed_ids = {
            work["id"]
            for work in self.project_ksg["schedule"]["works"]
            if work.get("name") in names_to_remove
        }

        # Remove nodes with specified names from the wg graph
        self.project_ksg["wg"]["nodes"] = [
            node
            for node in self.project_ksg["wg"]["nodes"]
            if node.get("work_unit", {}).get("name") not in names_to_remove
        ]

        # Update parent_edges in the remaining nodes to remove references to deleted works
        for node in self.project_ksg["wg"]["nodes"]:
            if "parent_edges" in node["work_unit"]:
                node["work_unit"]["parent_edges"] = [
                    edge
                    for edge in node["work_unit"]["parent_edges"]
                    if edge["str"] not in removed_ids
                ]

    def _trim_plan_to_n_works(self, n):
        if n >= len(self.project_ksg["schedule"]["works"]):
            return  # No trimming needed as n is greater than or equal to the number of works

        # Step 1: Identify the IDs of the first N works to keep
        keep_ids = set(work["id"] for work in self.project_ksg["schedule"]["works"][:n])

        # Step 2: Keep only the first N works in the schedule
        self.project_ksg["schedule"]["works"] = self.project_ksg["schedule"]["works"][
            :n
        ]

        # Step 3: Keep only the first N nodes in the wg graph
        self.project_ksg["wg"]["nodes"] = [
            node
            for node in self.project_ksg["wg"]["nodes"]
            if node["work_unit"]["id"] in keep_ids
        ]

        # Step 4: Update parent_edges in the remaining nodes
        for node in self.project_ksg["wg"]["nodes"]:
            if "parent_edges" in node["work_unit"]:
                node["work_unit"]["parent_edges"] = [
                    edge
                    for edge in node["work_unit"]["parent_edges"]
                    if edge["str"] in keep_ids
                ]



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .AbstractChecker import AbstractChecker
from .JournalChecker import JournalChecker


class TimeChecker(AbstractChecker):
    def validate(
        self, df_wind_model: pd.DataFrame, df_wind_val: pd.DataFrame, act: list
    ):
        df_wind_model = df_wind_model.drop_duplicates()
        df_stat = pd.DataFrame()
        dict_fig = {}
        final_df = pd.DataFrame()
        j = 0
        for c in act:
            dict_fig[c] = []
            for i in df_wind_model.index:
                if df_wind_model.loc[i, c] != 0:
                    c_act = [c]
                    volume = [df_wind_model.loc[i, ci] for ci in c_act]
                    delta = [self.percent_delta * volume_i for volume_i in volume]
                    sample = df_wind_val.copy()
                    for k, ci in enumerate(c_act):
                        sample = sample.loc[
                            (sample[ci] >= volume[k] - delta[k])
                            & (sample[ci] <= volume[k] + delta[k])
                        ]
                    if sample.shape[0] > 3:
                        df_stat, dict_for_sample = self.handle_sample(
                            i,
                            c,
                            df_wind_model,
                            sample,
                            df_wind_val,
                            df_stat,
                            j,
                        )
                        dict_fig[c].append(
                            {
                                "volume": volume,
                                "fig_data": dict_for_sample,
                                "color": dict_for_sample["Color"],
                            }
                        )
                    else:
                        journal_validation = JournalChecker()
                        color, q1, q99, prod = journal_validation.validate_time(
                            c_act, df_wind_model, i
                        )
                        if color == "grey":
                            df_stat.loc[j, "Work"] = c
                            df_stat.loc[j, "Time label"] = color
                        else:
                            (
                                blue_points,
                                black_points,
                                star,
                            ) = journal_validation.process_key_time(
                                c_act, df_wind_model, i, prod
                            )
                            fig_data = {
                                "Blue points": blue_points,
                                "Black points": black_points,
                                "Star": star,
                                "Color": color,
                                "Q1": q1,
                                "Q99": q99,
                            }
                            dict_fig[c].append(
                                {
                                    "volume": volume,
                                    "fig_data": fig_data,
                                    "color": color,
                                }
                            )
                            df_stat.loc[j, "Work"] = c
                            df_stat.loc[j, "Time label"] = color

                    j += 1

        not_grey = df_stat.loc[df_stat["Time label"] != "grey"]
        not_perc = ((df_stat.shape[0] - not_grey.shape[0]) / df_stat.shape[0]) * 100
        norm_df = df_stat.loc[df_stat["Time label"] == "green"]
        norm_perc = 0
        if not_perc != 100:
            norm_perc = ((norm_df.shape[0]) / not_grey.shape[0]) * 100

        final_df = self.finalize_dataframe(act, final_df, not_grey)
        return final_df, dict_fig, norm_perc, not_perc

    def handle_sample(self, i, c, df_wind_model, sample, df_wind_val, df_stat, j):
        value = df_wind_model.loc[i, c.split("_act_fact")[0] + "_real_time_act"]
        q1, q99 = np.quantile(
            sample[c.split("_act_fact")[0] + "_real_time_act"].values,
            [self.lower_quantile, self.upper_quantile],
        )
        q1 = int(q1)
        q99 = int(q99)
        if value < q1 or value > q99:
            color = "red"
        else:
            color = "green"
        df_stat.loc[j, "Work"] = c
        df_stat.loc[j, "Time label"] = color
        sample_dict = self.create_figure_dict(
            c, color, sample, df_wind_val, df_wind_model, i, q1, q99
        )
        return df_stat, sample_dict

    @staticmethod
    def create_figure_dict(
        c,
        color,
        sample,
        df_wind_val: pd.DataFrame,
        df_wind_model: pd.DataFrame,
        i,
        q1,
        q99,
    ):
        blue_points = {
            "x": list(df_wind_val[c].values),
            "y": list(df_wind_val[c.split("_act_fact")[0] + "_real_time_act"].values),
        }
        black_points = {
            "x": list(sample[c].values),
            "y": list(sample[c.split("_act_fact")[0] + "_real_time_act"].values),
        }
        star = {
            "x": df_wind_model.loc[i, c],
            "y": df_wind_model.loc[i, c.split("_act_fact")[0] + "_real_time_act"],
        }
        return {
            "Blue points": blue_points,
            "Black points": black_points,
            "Star": star,
            "Color": color,
            "Q1": q1,
            "Q99": q99,
        }

    @staticmethod
    def finalize_dataframe(act: list, final_df, not_grey):
        for i, c in enumerate(act):
            final_df.loc[i, "Name"] = c
            sample = not_grey.loc[not_grey["Work"] == c]
            count_dict = sample["Time label"].value_counts().to_dict()
            if "red" in count_dict:
                final_df.loc[i, "Time per Volume Unit"] = 0
            elif "green" not in count_dict and "red" not in count_dict:
                final_df.loc[i, "Time per Volume Unit"] = None
            else:
                final_df.loc[i, "Time per Volume Unit"] = (
                    count_dict["green"] / sample.shape[0]
                ) * 100
        return final_df

    def common_validation(self, ksg_data: dict, val_data, plan_type: str):
        df_validation_table = pd.DataFrame()
        fig_dict = dict()
        norm_perc = 0
        not_perc = 0
        i = 0
        if plan_type == "gpn":
            for w in ksg_data["schedule"]["works"]:
                i += 1
                work_name = w["display_name"]
                time = w["start_end_time"][1]["value"] - w["start_end_time"][0]["value"]
                work_vol = w["volume"]
                down_v = work_vol - 0.5 * work_vol
                up_v = work_vol + 0.5 * work_vol
                val_sample = val_data.loc[
                    (val_data["Work"] == work_name)
                    & (val_data["Volume"] >= down_v)
                    & (val_data["Volume"] <= up_v)
                ]
                if val_sample.shape[0] > 3:
                    q1, q99 = np.quantile(
                        val_sample["Time"].values,
                        [self.lower_quantile, self.upper_quantile],
                    )
                    if (time > q1) and (time < q99):
                        df_validation_table.loc[i, "Work name"] = work_name
                        df_validation_table.loc[i, "Validation Results"] = "green"
                        key = work_name
                        counts, bins, _ = plt.hist(val_sample["Time"].values)
                        fig_dict[key] = {
                            "Line": time,
                            "color": "green",
                            "Hight": counts,
                            "Bins": bins,
                            "Q1": q1,
                            "Q99": q99,
                        }
                    else:
                        df_validation_table.loc[i, "Work name"] = work_name
                        df_validation_table.loc[i, "Validation Results"] = "red"
                        key = work_name
                        counts, bins, _ = plt.hist(val_sample["Time"].values)
                        fig_dict[key] = {
                            "Line": time,
                            "color": "red",
                            "Hight": counts,
                            "Bins": bins,
                            "Q1": q1,
                            "Q99": q99,
                        }
                else:
                    df_validation_table.loc[i, "Work name"] = work_name
                    df_validation_table.loc[i, "Validation Results"] = None
        elif plan_type == "s7":
            for w in ksg_data["schedule"]["works"]:
                i += 1
                work_name = w["display_name"]
                time = w["start_end_time"][1]["value"] - w["start_end_time"][0]["value"]
                res_type = "res_parent"
                act_name = None
                defect_name = None
                if ":" in work_name:
                    act_name = work_name.split(":")[0]
                    defect_name = work_name.split(":")[-1]
                    res_type = "res_child"

                if ":" in work_name:
                    val_sample = val_data[1].loc[
                        (val_data[1]["class_child"] == defect_name)
                        & (val_data[1]["event_name_parent"] == act_name)
                    ]
                else:
                    val_sample = val_data[0].loc[
                        (val_data[0]["event_name_parent"] == work_name)
                    ]
                if val_sample.shape[0] > 3:
                    val_sample["Time"] = (
                        val_sample[res_type + "_users"]
                        / val_sample[res_type + "_hours"]
                    )
                    q1, q99 = np.quantile(
                        val_sample["Time"].values,
                        [self.lower_quantile, self.upper_quantile],
                    )
                    if (time >= q1) and (time <= q99):
                        df_validation_table.loc[i, "Work name"] = work_name
                        df_validation_table.loc[i, "Validation Results"] = "green"
                        key = work_name
                        counts, bins, _ = plt.hist(val_sample["Time"].values)
                        fig_dict[key] = {
                            "Line": time,
                            "color": "green",
                            "Hight": counts,
                            "Bins": bins,
                            "Q1": q1,
                            "Q99": q99,
                        }
                    else:
                        df_validation_table.loc[i, "Work name"] = work_name
                        df_validation_table.loc[i, "Validation Results"] = "red"
                        key = work_name
                        counts, bins, _ = plt.hist(val_sample["Time"].values)
                        fig_dict[key] = {
                            "Line": time,
                            "color": "red",
                            "Hight": counts,
                            "Bins": bins,
                            "Q1": q1,
                            "Q99": q99,
                        }
                else:
                    df_validation_table.loc[i, "Work name"] = work_name
                    df_validation_table.loc[i, "Validation Results"] = None
        not_perc = (
            (
                df_validation_table.loc[
                    df_validation_table["Validation Results"] == None
                ].shape[0]
            )
            / df_validation_table.shape[0]
        ) * 100
        if not_perc == 100:
            norm_perc = 0
        else:
            not_non = df_validation_table.loc[
                df_validation_table["Validation Results"] != None
            ]
            norm_perc = (
                (not_non.loc[not_non["Validation Results"] == "green"].shape[0])
                / not_non.shape[0]
            ) * 100
        return df_validation_table, fig_dict, norm_perc, not_perc



import datetime
import json
from copy import copy
from os import listdir
from os.path import isfile, join

import pandas as pd
import pkg_resources
from rich.console import Console
from rich.progress import Progress

from stairsval.core.aggregator import Aggregator
from stairsval.core.database.DBWrapper import DBWrapper
from stairsval.core.dataset_processors.KSGDataset import KSGDataset
from stairsval.core.dataset_processors.PlanDataset import PlanDataset
from stairsval.core.dataset_processors.ValidationDataset import ValidationDataset
from stairsval.core.expert_metrics_estimator import ExpertMetricsEstimator
from stairsval.core.stats_calculator import StatisticCalculator
from stairsval.validation_checks.ResourcesChecker import ResourcesChecker
from stairsval.validation_checks.TimeChecker import TimeChecker
from stairsval.validation_checks.WorksChecker import WorksChecker
from stairsval.validators.BaseValidator import BaseValidator

brave_precalculated = "../tables/precalculated_brave_coefficients.csv"
links_history = "../tables/links_history_messoyaha_with_granular_smr_v4.csv"
brave_path = pkg_resources.resource_filename(__name__, brave_precalculated)
links_history_path = pkg_resources.resource_filename(__name__, links_history)


class GPNPlanValidator:
    def __init__(self, plan, connector=None, level=None):
        self.console = Console()
        self.ksg_data = KSGDataset(
            ksg_data=plan, connector=connector, level=level
        ).collect()
        self.plan_dataset = PlanDataset(ksg_data=self.ksg_data)
        self.data_source = DBWrapper(connector=connector, level=level)
        self.aggregator = Aggregator()
        self.works_validator = WorksChecker()
        self.time_validator = TimeChecker()
        self.resources_validator = None
        self.statistic_calculator = None

    def validate(self):
        with Progress(console=self.console) as progress:
            validation_task = progress.add_task("[cyan]Validating...", total=9)

            self.console.log("Collecting data")
            plan_df = self.plan_dataset.collect()
            act = self.plan_dataset.get_act_names()
            res = self.plan_dataset.get_res_names()
            brave = pd.read_csv(brave_path, index_col=0)

            pools = self.plan_dataset.get_pools()

            validation_df = ValidationDataset(
                self.data_source, pools, act, res
            ).collect()
            progress.advance(validation_task)

            self.console.log("Validating resources on log data")

            self.resources_validator = ResourcesChecker(
                res=[r + "_res_fact" for r in res]
            )
            (
                df_perc_agg_res,
                df_final_volume_res,
                df_final_style_res,
                fig_dict_res,
                not_perc_res,
                norm_perc_res,
            ) = self.resources_validator.validate(plan_df, validation_df, act)
            progress.advance(validation_task)

            self.console.log("Validating resources on journal data")
            (
                df_vedom,
                not_perc_vedom,
                norm_perc_vedom,
            ) = self.aggregator.get_res_ved_stat(brave, self.ksg_data, plan_type="gpn")
            progress.advance(validation_task)

            self.console.log("Splitting data into window frames")
            _, df_wind_val = self.window_zero(validation_df)
            _, df_wind_model = self.window_zero(plan_df)
            for i in df_wind_val.index:
                for c in act:
                    df_wind_val.loc[i, c.split("_")[0] + "_real_time_act"] = (
                        df_wind_val.loc[i, "Window length"]
                        - df_wind_val.loc[i, c.split("_")[0] + "_act_fact_zero_lil"]
                        - df_wind_val.loc[i, c.split("_")[0] + "_act_fact_zero_big"]
                    )
            for i in df_wind_model.index:
                for c in act:
                    df_wind_model.loc[i, c.split("_")[0] + "_real_time_act"] = (
                        df_wind_model.loc[i, "Window length"]
                        - df_wind_model.loc[i, c.split("_")[0] + "_act_fact_zero_lil"]
                        - df_wind_model.loc[i, c.split("_")[0] + "_act_fact_zero_big"]
                    )
            progress.advance(validation_task)

            self.console.log("Validation of average daily work volumes")
            (
                df_volume_stat,
                dist_dict,
                norm_volume_perc,
                not_volume_perc,
            ) = self.works_validator.validate(
                plan_df, validation_df, act=[c + "_act_fact" for c in act]
            )
            progress.advance(validation_task)

            self.console.log("Validating time")
            (
                df_time_stat,
                time_plot_dict,
                norm_time_perc,
                not_time_perc,
            ) = self.time_validator.validate(
                df_wind_model, df_wind_val, act=[c + "_act_fact" for c in act]
            )
            progress.advance(validation_task)
            self.console.log("Validating sequences")
            seq_history = pd.read_csv(links_history_path)
            (
                df_seq_stat,
                df_color_seq,
                norm_perc_seq,
                not_perc_seq,
            ) = self.aggregator.get_all_seq_statistic(seq_history, self.ksg_data)
            progress.advance(validation_task)
            self.console.log(
                "Collecting general statistics on activities and resources"
            )
            self.statistic_calculator = StatisticCalculator(
                norm_perc_res,
                norm_perc_vedom,
                norm_volume_perc,
                norm_time_perc,
                norm_perc_seq,
                not_perc_res,
                not_perc_vedom,
                not_volume_perc,
                not_time_perc,
                not_perc_seq,
            )
            work_res_common_perc = (
                self.statistic_calculator.get_statistic_for_properties_and_all_stat()
            )
            plan_statistic = self.statistic_calculator.get_plan_statistic()
            progress.advance(validation_task)

        return (
            df_perc_agg_res,
            df_final_volume_res,
            df_final_style_res,
            fig_dict_res,
            df_vedom,
            df_volume_stat,
            dist_dict,
            df_time_stat,
            time_plot_dict,
            df_seq_stat,
            df_color_seq,
            work_res_common_perc,
            plan_statistic,
        )

    def common_validate(self):
        act = self.plan_dataset.get_act_names()
        res = self.plan_dataset.get_res_names()
        brave = pd.read_csv(brave_path, index_col=0)
        pools = self.plan_dataset.get_pools()
        validation_df_res = ValidationDataset(
            self.data_source, pools, act, res
        ).collect()
        val_data_time = self.data_source.get_time_data(act)
        (
            df_validation_table_res,
            fig_dict_res,
            norm_perc_res_val,
            not_perc_res_val,
        ) = ResourcesChecker(res=res).common_validation(
            self.ksg_data, validation_df_res, plan_type="gpn"
        )
        df_vedom, not_perc_vedom, norm_perc_vedom = self.aggregator.get_res_ved_stat(
            brave, self.ksg_data, plan_type="gpn"
        )
        (
            df_validation_table_time,
            fig_dict_time,
            norm_perc_time,
            not_perc_time,
        ) = TimeChecker().common_validation(self.ksg_data, val_data_time, "gpn")
        return (
            df_validation_table_res,
            fig_dict_res,
            norm_perc_res_val,
            not_perc_res_val,
            df_vedom,
            not_perc_vedom,
            norm_perc_vedom,
            df_validation_table_time,
            fig_dict_time,
            norm_perc_time,
            not_perc_time,
        )

    @staticmethod
    def window_zero(df: pd.DataFrame, min_window: int = 2, max_window: int = 10):
        df = df.drop_duplicates()
        base = datetime.datetime.today()
        date_list = [
            (base + datetime.timedelta(days=x)).strftime("%d.%m.%Y")
            for x in range(df.shape[0])
        ]
        df.index = date_list
        """Function for making time windows of msg files

        Args:
            df (pd.DataFrame): input msg file
            min_window (int, optional): Min length of window. Defaults to 5.
            max_window (int, optional): Max length of window. Defaults to 31.

        Returns:
            DataFrame: msg files with windows
        """
        act_col = [var for var in df.columns if "_act_fact" in var]

        res_col = [var for var in df.columns if "_res_fact" in var]
        new_zero_lil = [act + "_zero_lil" for act in act_col]
        new_zero_big = [act + "_zero_big" for act in act_col]
        df_new = pd.DataFrame(
            columns=list(df.columns)
            + ["Start day", "Window length"]
            + new_zero_lil
            + new_zero_big
        )
        df_zero = pd.DataFrame(columns=act_col)
        for window_len in [7]:
            for start in range(0, len(df) - window_len + 1):
                new_row = (
                    df[act_col]
                    .iloc[start : start + window_len]
                    .sum(axis=0, skipna=True)
                )
                new_row = {
                    act: 1 if value == 0.0 else 0 for act, value in new_row.items()
                }
                start_day = datetime.datetime.strptime(df.index[start], "%d.%m.%Y")
                end_day = datetime.datetime.strptime(
                    df.index[start + window_len - 1], "%d.%m.%Y"
                )
                df_zero.loc[df.index[start]] = new_row

        window_list = list(range(min_window, max_window + 1))
        for window_len in window_list:
            for start in range(0, len(df) - window_len + 1):
                start_day = datetime.datetime.strptime(df.index[start], "%d.%m.%Y")
                end_day = datetime.datetime.strptime(
                    df.index[start + window_len - 1], "%d.%m.%Y"
                )
                if (end_day - start_day).days == window_len - 1:
                    new_row_act = (
                        df[act_col]
                        .iloc[start : start + window_len]
                        .sum(axis=0, skipna=True)
                    )
                    new_row_res = (
                        df[res_col]
                        .iloc[start : start + window_len]
                        .mean(axis=0, skipna=True)
                    )
                    new_row = copy(new_row_act)
                    new_row = pd.concat([new_row, new_row_res])

                    lil = {act + "_zero_lil": 0 for act in act_col}
                    big = {act + "_zero_big": 0 for act in act_col}
                    danger_date = [
                        df.index[ind] for ind in range(start, start + window_len)
                    ]

                    for act in act_col:
                        for i in range(start, start + window_len):
                            if df.iloc[i][act] == 0.0:
                                lil[act + "_zero_lil"] = lil[act + "_zero_lil"] + 1
                                if (
                                    (i < window_len - 6)
                                    and danger_date[i - start] in df_zero.index
                                    and df_zero.loc[danger_date[i - start]][act] == 1
                                ):
                                    if (i > start) and (
                                        df_zero.loc[danger_date[i - start - 1]][act]
                                        == 1
                                    ):
                                        big[act + "_zero_big"] = (
                                            big[act + "_zero_big"] + 1
                                        )
                                    else:
                                        big[act + "_zero_big"] = (
                                            big[act + "_zero_big"] + 7
                                        )
                    lil = {
                        act
                        + "_zero_lil": lil[act + "_zero_lil"]
                        - big[act + "_zero_big"]
                        for act in act_col
                    }

                    new_row = {
                        **new_row,
                        **lil,
                        **big,
                        "Start day": start_day,
                        "Window length": window_len,
                    }
                    df_new = pd.concat(
                        [df_new, pd.DataFrame(pd.Series(new_row)).transpose()],
                        ignore_index=True,
                    )

        return df_zero, df_new

    @staticmethod
    def _file_opener(mypath, validation_files, act):
        onlyfiles = [f for f in listdir(mypath) if isfile(join(mypath, f))]
        for file in onlyfiles:
            msg_file = open(mypath + file, encoding="utf8")
            msg_data = json.load(msg_file)
            if len(msg_data["resource"]) != 0:
                for work in msg_data["work"]:
                    if work["work title"] in act:
                        validation_files.append(mypath + file)
                        break
        return validation_files

    @staticmethod
    def _save_json_file(label, json_file, new_plan_path):
        with open(new_plan_path + str(label) + ".json", "w") as outfile:
            json.dump(json_file, outfile)
        return new_plan_path + str(label) + ".json"


class GPNValidator(BaseValidator):
    """Validator for GPN plans."""

    def __init__(self, project_ksg, connector=None, level=None):
        super().__init__(project_ksg)
        self.connector = connector
        self.level = level
        self.plan_type = "gpn"

    def specific_validation(self):
        result_dict = dict()
        ksg_for_val_data = self.project_ksg
        regime = True
        for el in ksg_for_val_data["schedule"]["works"]:
            if (
                el["display_name"] == "Окончание работ по марке"
                or el["display_name"] == "Начало работ по марке"
            ):
                regime = False
                break
        if regime:
            (
                df_perc_agg_res,
                df_final_volume_res,
                df_final_style_res,
                fig_dict_res,
                df_vedom,
                df_volume_stat,
                dist_dict,
                df_time_stat,
                time_plot_dict,
                df_seq_stat,
                df_color_seq,
                work_res_common_perc,
                plan_statistic,
            ) = GPNPlanValidator(
                plan=self.project_ksg,
                connector=self.connector,
                level=self.level,
            ).validate()

            result_dict["Object 0"] = {
                "Resource to Work Ratio": df_perc_agg_res,
                "Work Pools Volumes": df_final_volume_res,
                "Work Pools Colors": df_final_style_res,
                "Data for Resource Charts": fig_dict_res,
                "Resource Journal": df_vedom,
                "Production Statistics": df_volume_stat,
                "Data for Production Charts": dist_dict,
                "Work Time Statistics": df_time_stat,
                "Data for Time Charts": time_plot_dict,
                "Connection Table": df_seq_stat,
                "Color Connection Table": df_color_seq,
                "Final Statistics for Work and Resources": work_res_common_perc,
                "Final Plan Statistics": plan_statistic,
            }
        else:
            json_files = []
            act_dict = dict()
            block_id = dict()
            ind_act = dict()
            label = 0
            for i, el in enumerate(ksg_for_val_data["schedule"]["works"]):
                act_dict[el["id"]] = el["display_name"]
                if el["display_name"] == "Окончание работ по марке":
                    label += 1
                elif el["display_name"] == "Начало работ по марке":
                    label += 1
                else:
                    block_id[el["id"]] = label
                    ind_act[el["id"]] = i
            block_id["0000000"] = label + 1
            json_file = {"schedule": {"works": []}, "wg": {"nodes": []}}
            label = block_id[list(block_id.keys())[0]]
            for k in block_id.keys():
                if block_id[k] == label:
                    ind_of_act = ind_act[k]
                    new_des_act = []
                    for el in ksg_for_val_data["wg"]["nodes"][ind_of_act][
                        "parent_edges"
                    ]:
                        if el[0] in block_id.keys():
                            if (
                                act_dict[el[0]] != "Окончание работ по марке"
                                and act_dict[el[0]] != "Начало работ по марке"
                            ):
                                if block_id[k] == block_id[el[0]]:
                                    new_des_act.append(el)
                    ksg_for_val_data["wg"]["nodes"][ind_of_act][
                        "parent_edges"
                    ] = new_des_act
                    json_file["schedule"]["works"].append(
                        ksg_for_val_data["schedule"]["works"][ind_of_act]
                    )
                    json_file["wg"]["nodes"].append(
                        ksg_for_val_data["wg"]["nodes"][ind_of_act]
                    )
                    label = block_id[k]
                else:
                    json_files.append(json_file)
                    if k == "0000000":
                        break
                    else:
                        json_file = {"schedule": {"works": []}, "wg": {"nodes": []}}
                        ind_of_act = ind_act[k]
                        new_des_act = []
                        for el in ksg_for_val_data["wg"]["nodes"][ind_of_act][
                            "parent_edges"
                        ]:
                            if el[0] in block_id.keys():
                                if (
                                    act_dict[el[0]] != "Окончание работ по марке"
                                    and act_dict[el[0]] != "Начало работ по марке"
                                ):
                                    if block_id[k] == block_id[el[0]]:
                                        new_des_act.append(el)
                        ksg_for_val_data["wg"]["nodes"][ind_of_act][
                            "parent_edges"
                        ] = new_des_act
                        json_file["schedule"]["works"].append(
                            ksg_for_val_data["schedule"]["works"][ind_of_act]
                        )
                        json_file["wg"]["nodes"].append(
                            ksg_for_val_data["wg"]["nodes"][ind_of_act]
                        )
                        label = block_id[k]
            for id_number, p in enumerate(json_files):
                plan = KSGDataset(
                    ksg_data=p, connector=self.connector, level=self.level
                ).collect()
                if len(plan["schedule"]["works"]) == 0:
                    continue
                else:
                    (
                        df_perc_agg_res,
                        df_final_volume_res,
                        df_final_style_res,
                        fig_dict_res,
                        df_vedom,
                        df_volume_stat,
                        dist_dict,
                        df_time_stat,
                        time_plot_dict,
                        df_seq_stat,
                        df_color_seq,
                        work_res_common_perc,
                        plan_statistic,
                    ) = GPNPlanValidator(
                        plan=p, connector=self.connector, level=self.level
                    ).validate()

                    result_dict["Object " + str(id_number)] = {
                        "Resource to Work Ratio": df_perc_agg_res,
                        "Work Pools Volumes": df_final_volume_res,
                        "Work Pools Colors": df_final_style_res,
                        "Data for Resource Charts": fig_dict_res,
                        "Resource Journal": df_vedom,
                        "Production Statistics": df_volume_stat,
                        "Data for Production Charts": dist_dict,
                        "Work Time Statistics": df_time_stat,
                        "Data for Time Charts": time_plot_dict,
                        "Connection Table": df_seq_stat,
                        "Color Connection Table": df_color_seq,
                        "Final Statistics for Work and Resources": work_res_common_perc,
                        "Final Plan Statistics": plan_statistic,
                    }
        return result_dict

    def calculate_expert_metrics(self):
        metrics_calculator = ExpertMetricsEstimator(self.project_ksg)
        metrics = metrics_calculator.calculate_metrics()
        formal_metrics = metrics_calculator.calculate_formal_metrics()
        return metrics, formal_metrics

    @staticmethod
    def _filter_activities(k, block_id, act_dict, ksg_for_val_data, ind_act):
        ind_of_act = ind_act[k]
        new_des_act = []
        for el in ksg_for_val_data["wg"]["nodes"][ind_of_act]["parent_edges"]:
            if el[0] in block_id.keys():
                if act_dict[el[0]] not in [
                    "Окончание работ по марке",
                    "Начало работ по марке",
                ]:
                    if block_id[k] == block_id[el[0]]:
                        new_des_act.append(el)
        ksg_for_val_data["wg"]["nodes"][ind_of_act]["parent_edges"] = new_des_act
        return ksg_for_val_data["schedule"]["works"][ind_of_act]

    def common_validation(self, cut_to_n_works: int = None):
        if cut_to_n_works:
            self._trim_plan_to_n_works(cut_to_n_works)

        (
            df_validation_table_res,
            fig_dict_res,
            norm_perc_res_val,
            not_perc_res_val,
            df_vedom,
            not_perc_vedom,
            norm_perc_vedom,
            df_validation_table_time,
            fig_dict_time,
            norm_perc_time,
            not_perc_time,
        ) = GPNPlanValidator(
            plan=self.project_ksg,
            connector=self.connector,
            level=self.level,
        ).common_validate()
        work_res_stat = dict()
        work_res_stat["Percentage of Normal Resource Volume Values"] = round(
            norm_perc_res_val
        )
        work_res_stat["Percentage of Atypical Resource Volume Values"] = 100 - round(
            norm_perc_res_val
        )
        work_res_stat["Percentage of Normal Resources According to Statements"] = round(
            norm_perc_vedom
        )
        work_res_stat[
            "Percentage of Atypical Resources According to Statements"
        ] = 100 - round(norm_perc_vedom)
        work_res_stat["Percentage of Normal Values Across All Resources"] = round(
            (
                work_res_stat["Percentage of Normal Resource Volume Values"]
                + work_res_stat[
                    "Percentage of Normal Resources According to Statements"
                ]
            )
            / 2
        )
        work_res_stat["Percentage of Atypical Values Across All Resources"] = (
            100 - work_res_stat["Percentage of Normal Values Across All Resources"]
        )
        work_res_stat["Percentage of Normal Work Time Values"] = round(norm_perc_time)
        work_res_stat["Percentage of Atypical Work Time Values"] = 100 - round(
            norm_perc_time
        )
        work_res_stat["Percentage of Normal Values Across All Works"] = round(
            norm_perc_time
        )
        work_res_stat["Percentage of Atypical Values Across All Works"] = 100 - round(
            norm_perc_time
        )
        work_res_stat["Percentage of Normal Plan Values"] = round(
            (
                work_res_stat["Percentage of Normal Values Across All Resources"]
                + work_res_stat["Percentage of Normal Values Across All Works"]
            )
            / 2
        )
        work_res_stat["Percentage of Atypical Plan Values"] = (
            100 - work_res_stat["Percentage of Normal Plan Values"]
        )
        work_res_stat["Percentage of Plan Values Not Covered by Validation"] = round(
            (not_perc_res_val + not_perc_vedom + not_perc_time) / 3
        )
        work_res_stat["Percentage of Plan Values Covered by Validation"] = 100 - round(
            (not_perc_res_val + not_perc_vedom + not_perc_time) / 3
        )

        result_dict = dict()
        result_dict["Common validation"] = {
            "Resources Validation Table": df_validation_table_res,
            "Data for Resource Charts": fig_dict_res,
            "Resource Journal": df_vedom,
            "Work Time Validation Table": df_validation_table_time,
            "Data for Time Charts": fig_dict_time,
            "Final Plan Statistics": work_res_stat,
        }
        return result_dict



import pandas as pd

from stairsval.core.database.Adapter import DBAdapter
from stairsval.core.database.DataSource import DataSource


class DBWrapper(DataSource):
    def __init__(self, connector=None, level=None):
        # Default value for connector if not provided
        if connector is None:
            connector = DBAdapter
        # Default value for level if not provided
        if level is None:
            level = connector.GRANULARY

        self.adapter = connector
        self.level = level

    def get_data(self, pools, res_names):
        validation_dataset_list = []
        for df in self.adapter.get_works_by_pulls(
            work_pulls=pools,
            resource_list=res_names,
            key=self.level,
            res_key=self.adapter.GRANULARY,
        ):
            if df is not None:
                df.fillna(0, inplace=True)
            validation_dataset_list.append(df)
        return validation_dataset_list

    def get_act_names(self):
        df = self.adapter.get_all_works_name()
        return df

    def get_res_names(self):
        df = self.adapter.get_all_resources_name()
        return df

    def get_time_data(self, acts):
        frames = []
        for p in self.adapter.from_names(
            works=acts, key=self.level, objects_limit=-1, ceil_limit=-1
        ):
            frames.append(p)
        validation_dataset = pd.DataFrame()
        for df in frames:
            validation_dataset = pd.concat([validation_dataset, df])
        validation_dataset = validation_dataset[
            ["granulary_name", "physical_volume", "start_date", "finish_date"]
        ]
        validation_dataset.dropna(inplace=True)
        validation_dataset = validation_dataset.loc[
            validation_dataset["granulary_name"].isin(acts)
        ]
        validation_dataset.reset_index(inplace=True, drop=True)
        validation_dataset[["start_date", "finish_date"]] = validation_dataset[
            ["start_date", "finish_date"]
        ].apply(pd.to_datetime)
        validation_dataset["Time"] = (
            validation_dataset["finish_date"] - validation_dataset["start_date"]
        )
        validation_dataset["Time"] = validation_dataset["Time"].dt.days
        validation_dataset = validation_dataset.drop(
            columns=["start_date", "finish_date"]
        )
        validation_dataset.columns = ["Work", "Volume", "Time"]
        return validation_dataset



from stairsval.core.dataset_processors.BaseDataset import BaseDataset


class S7KSG(BaseDataset):
    def __init__(self, ksg_data):
        self.ksg_data = ksg_data

    def collect(self):
        new_act = []
        new_wg = []
        deleted_id = []
        for i, w in enumerate(self.ksg_data["schedule"]["works"]):
            if float(w["volume"]) != 0.0:
                new_act.append(w)
                new_wg.append(self.ksg_data["wg"]["nodes"][i])
            else:
                deleted_id.append(w["id"])
        self.ksg_data["schedule"]["works"] = new_act
        self.ksg_data["wg"]["nodes"] = new_wg
        for w in self.ksg_data["wg"]["nodes"]:
            new_des = []
            for el in w["parent_edges"]:
                if el[0] not in deleted_id:
                    new_des.append(el)
            w["parent_edges"] = new_des
        return self.ksg_data



from typing import Generator

import pandas as pd


class DBAdapter:
    # name level constants
    GRANULARY: dict[str, str] = {}
    PROCESSED: dict[str, str] = {}
    TYPEDLVL2: dict[str, str] = {}

    def __init__(self, *args, **kwargs):
        """

        Args:
            connect: connect object for database.
                     can be a reference to a database service, as well as a sql connect object or None

        """
        raise NotImplementedError

    def from_names(
        self,
        works: list[str],
        resources: list[str] = (),
        ceil_limit: int = 1_000,
        objects_limit: int = 1,
        crossing: bool = False,
        key: dict[str, str] = GRANULARY,
        res_key: dict[str, str] = GRANULARY,
        *args,
        **kwargs,
    ) -> Generator[pd.DataFrame, None, None]:
        """
        Returns all schedules that include jobs and resources from the lists

        Args:
            works: list of works for which schedules are searched
            resources: list of resources for which schedules are searched
            ceil_limit: limit of records for one-time issuance from the database
            objects_limit: limit of objects for one-time issuance from the database
            crossing: flag describing the logic of combining resources and activities.
                      true - select schedules than contain work AND resources
                      false - select schedules than contain work OR resources
            key: level of works names
            res_key: level of resources names

        Returns:
            Pandas dataframe generator
        """
        raise NotImplementedError

    def get_works_by_pulls(
        self,
        work_pulls: list[list[str]],
        resource_list: list[list[str]] = [],
        key: dict[str, str] = GRANULARY,
        res_key: dict[str, str] = GRANULARY,
        *args,
        **kwargs,
    ) -> Generator[pd.DataFrame, None, None]:
        """
        Generator
        Returns information about work pools

        Args:
            work_pulls: list of works pulls
            resource_list: list of resources using on pulls
            key: level of works names
            res_key: level of resources names

        Returns:
            Pandas dataframe generator
        """
        raise NotImplementedError

    def get_all_works_name(self, *args, **kwargs) -> pd.DataFrame:
        """
        Returns all works names

        Returns:
            pandas dataframe
        """
        raise NotImplementedError

    def get_all_resources_name(self, *args, **kwargs) -> pd.DataFrame:
        """
        Returns all resources names

        Returns:
            pandas dataframe

        """
        raise NotImplementedError
